import { HttpService, Injectable } from '@nestjs/common';
import { Logger } from '@nestjs/common';
import { Observable } from 'rxjs';
import { AxiosResponse } from 'axios';
import { Cron, CronExpression } from '@nestjs/schedule';
import * as _ from 'lodash';
import * as https from 'https';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { ApticStructure } from '../schemas/aptic-structure.schema';
import { Address } from '../schemas/address.schema';
import { UsersService } from '../../users/users.service';

@Injectable()
export class ApticStructuresService {
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  public formatApticStructures(postalCodeData: any[]): any {
    // Get all postal code in one array
    const postalCodeArray = _.flatten(
      postalCodeData.map((data) => {
        return data.codesPostaux;
      })
    );

    // Call APTIC Api's
    const postalCodePromises = postalCodeArray.map((postalCode) => {
      return this.getApticStructures(postalCode).toPromise();
    });

    Promise.all(postalCodePromises).then((data) => {
      const structuresData = _.flatten(
        data.map((tmp: { data }) => {
          return tmp.data.data;
        })
      );
      // Create structures if possible
      structuresData.forEach((structure) => this.createApticStructures(structure));
    });
  }

  private async createApticStructures(structure: ApticStructure): Promise<any> {
    this.structureAlreadyExist(structure).then((exist) => {
      if (!exist) {
        Logger.log(`Create structure : ${structure.presence_name}`, 'ApticStructuresService - createApticStructures');
        const createdStructure = new this.structureModel();
        // Known fields
        createdStructure.structureName = structure.presence_name;
        createdStructure.contactPhone = structure.presence_phone;
        // Unkown fields (but mandatory)
        createdStructure.contactMail = 'unknown@unknown.com';
        createdStructure.labelsQualifications = ['passNumerique'];
        createdStructure.structureType = 'autre';
        createdStructure.pmrAccess = false;
        createdStructure.accessModality = ['accesLibre'];
        createdStructure.publics = ['toutPublic'];
        createdStructure.accountVerified = true;
        createdStructure.freeWorkShop = false;
        createdStructure.nbComputers = null;
        createdStructure.nbPrinters = null;
        createdStructure.nbScanners = null;
        createdStructure.nbTablets = null;
        createdStructure.nbNumericTerminal = null;
        // Address
        createdStructure.coord = [structure.gps_lng, structure.gps_lat];
        createdStructure.address = this.formatAddress(structure);
        createdStructure.save();
        // Send admin weird structure mail
        this.verifyDuplication(createdStructure);
      }
    });
  }

  private async structureAlreadyExist(structure: ApticStructure): Promise<boolean> {
    let existingStructure = await this.structureModel
      .findOne({
        structureName: { $regex: structure.presence_name, $options: 'i' },
      })
      .exec();
    // Check without regex for case like 'TINEBRA*DANIEL/DANIEL/'
    if (!existingStructure) {
      existingStructure = await this.structureModel.findOne({ structureName: structure.presence_name }).exec();
    }

    if (existingStructure) {
      // Add aptic label if it's not the case
      if (!existingStructure.labelsQualifications.includes('passNumerique')) {
        existingStructure.labelsQualifications.push('passNumerique');
        existingStructure.save();
      }
      return true;
    }
    return false;
  }

  /**
   * Get Metropole new aptic structure evey week. For testing, please change the expression
   */
  @Cron(CronExpression.EVERY_WEEK)
  public getMetopoleMunicipality(): void {
    const req =
      'https://download.data.grandlyon.com/ws/grandlyon/adr_voie_lieu.adrcomgl/all.json?maxfeatures=-1&start=1';
    Logger.log(`Request : ${req}`, 'ApticStructuresService - getMetopoleMunicipality');
    this.httpService.get(encodeURI(req)).subscribe(
      (data) => {
        const inseeArray = data.data.values.map((municipality) => {
          return this.getPostalCodeWithINSEE(municipality.insee).toPromise();
        });
        Promise.all(inseeArray).then((inseData) => {
          const postalCodeArray = inseData.map((cpData: { data; config }) => {
            return cpData.data;
          });
          this.formatApticStructures(postalCodeArray);
        });
      },
      (err) => Logger.error(err)
    );
  }

  public getPostalCodeWithINSEE(inseeCode: string): Observable<AxiosResponse<any>> {
    const req = `https://geo.api.gouv.fr/communes/${inseeCode}?fields=codesPostaux&format=json`;
    Logger.log(`Request : ${req}`, 'ApticStructuresService - getMetopoleMunicipality');
    return this.httpService.get(encodeURI(req));
  }

  public getApticStructures(postalCodeData: string): Observable<AxiosResponse<any>> {
    const agent = new https.Agent({
      rejectUnauthorized: false,
    });
    const req = `https://presence.aptic.fr/postal_code/${postalCodeData}`;
    Logger.log(`Request : ${req}`, 'ApticStructuresService');
    return this.httpService.get(req, {
      httpsAgent: agent,
      headers: {
        api_key: process.env.APTIC_TOKEN,
      },
    });
  }

  private async verifyDuplication(createdStructure: StructureDocument): Promise<void> {
    const sameAddrStructure = await this.structureModel
      .findOne({
        _id: { $ne: createdStructure._id },
        address: createdStructure.address,
      })
      .exec();
    if (sameAddrStructure) {
      this.userService.sendAdminApticStructureMail(createdStructure.structureName, sameAddrStructure.structureName);
    }
  }

  /**
   * Format aptic structure address
   */
  private formatAddress(structure: ApticStructure): Address {
    const address = new Address();
    const regexWithSpace = /\d+\s/g; // NOSONAR
    const regex = /\d+/g; // NOSONAR
    if (structure.presence_address.match(regex)) {
      address.numero = structure.presence_address.match(regex)[0];
      address.street = structure.presence_address.replace(regexWithSpace, '');
    } else {
      address.street = structure.presence_address;
    }
    address.commune = structure.city;
    return address;
  }
}
